import pyvista as pv


def extract_surface(img, resolution=(1, 1, 1), origin=(0, 0, 0)):
    # img should be a binary 3D np.array
    grid = pv.ImageData(dimensions=img.shape, spacing=resolution, origin=origin)
    mesh = grid.contour(isosurfaces=[0.5], scalars=img.flatten(order="F"), method="marching_cubes")
    surf = mesh.extract_geometry()
    surf.clear_data()
    return surf



import argparse
import typing

from . import volume_clip, volume_slice, mpl_slice, volume_threshold
from .mpl_slice import plot_slices

__all__ = ["add_viz_parser", "dispatch", "numpy_to_vtkImageData", "plot_slices", "mpl_slice"]


def list_viz_commands() -> list[str]:
    return ["volume-clip", "volume-slice", "volume-threshold", "mpl-slice"]


def add_viz_parser(parser: argparse.ArgumentParser) -> None:
    subparsers = parser.add_subparsers(dest="viz-command")

    volume_clip_parser = subparsers.add_parser("volume-clip", help="Clip a volume")
    volume_clip.add_arguments(volume_clip_parser)

    volume_slice_parser = subparsers.add_parser("volume-slice", help="Slice a volume")
    volume_slice.add_arguments(volume_slice_parser)

    volume_threshold_parser = subparsers.add_parser("volume-threshold", help="Threshold a volume")
    volume_threshold.add_arguments(volume_threshold_parser)

    mpl_slice_parser = subparsers.add_parser("mpl-slice", help="Plot slices using matplotlib")
    mpl_slice.add_arguments(mpl_slice_parser)


def dispatch(command, args: dict[str, typing.Any]) -> int:
    if command == "volume-clip":
        volume_clip.main(**args)

    elif command == "volume-slice":
        volume_slice.main(**args)

    elif command == "volume-threshold":
        volume_threshold.main(**args)

    elif command == "mpl-slice":
        mpl_slice.main(**args)

    else:
        raise ValueError(f"Unknown command {command}")

    return 0



import typing
import logging
import argparse
from pathlib import Path

from . import basic, idealized_brain
from .basic import create_mesh

__all__ = ["basic", "idealized_brain", "create_mesh"]

logger = logging.getLogger(__name__)


def add_mesh_parser(parser: argparse.ArgumentParser) -> None:
    subparsers = parser.add_subparsers(dest="mesh-command")

    template_parser = subparsers.add_parser(
        "template", help="Create a sample configuration file called"
    )
    template_parser.add_argument("outdir", type=Path, help="Output directory")
    template_parser.add_argument(
        "--name", type=str, default="config.json", help="Name of the configuration file"
    )

    create_parser = subparsers.add_parser("create", help="Create a mesh")
    create_parser.add_argument("filename", type=Path, help="Path to the configuration file")

    convert_parser = subparsers.add_parser("convert", help="Convert mesh to dolfinx")
    convert_parser.add_argument("mesh_dir", type=Path, help="Directory containing mesh files")

    idealized_parser = subparsers.add_parser(
        "idealized",
        help="Generate idealized surface",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    idealized_brain.add_arguments(idealized_parser)


def dispatch(command, args: dict[str, typing.Any]) -> int:
    if command == "template":
        basic.generate_sameple_config(**args)

    elif command == "create":
        mesh_dir = basic.create_mesh_from_config(**args)
        try:
            basic.convert_mesh_dolfinx(mesh_dir=mesh_dir)
        except ImportError:
            logger.debug("dolfinx not installed, skipping conversion to dolfinx")

    elif command == "convert":
        basic.convert_mesh_dolfinx(**args)

    elif command == "idealized":
        idealized_brain.main(**args)

    else:
        raise ValueError(f"Unknown command {command}")

    return 0



from __future__ import annotations
import argparse
import typing
from pathlib import Path
import numpy as np


def add_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-i",
        "--input",
        type=Path,
        required=True,
        help="Path to the file to analyze",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        default="results",
        help="Path to the output directory",
    )
    parser.add_argument(
        "--tag",
        "-t",
        type=str,
        default="",
        help="Tag to add to the output files",
    )
    parser.add_argument(
        "--cmap",
        type=str,
        default="gray",
        help="Colormap to use",
    )
    parser.add_argument(
        "--add-colorbar",
        action="store_true",
        help="Add colorbar to the plots",
    )
    parser.add_argument(
        "--nx",
        type=int,
        default=5,
        help="Number of rows in the plot",
    )
    parser.add_argument(
        "--ny",
        type=int,
        default=5,
        help="Number of columns in the plot",
    )


def main(
    input: Path, output: Path, tag: str, cmap: str, add_colorbar: bool, nx: int, ny: int
) -> int:
    output.mkdir(parents=True, exist_ok=True)
    import nibabel as nib

    img = nib.load(input).get_fdata()
    plot_slices(
        img,
        output,
        tag=tag,
        cmap=cmap,
        add_colorbar=add_colorbar,
        nx=nx,
        ny=ny,
    )
    return 0


def plot_slice(
    img: np.ndarray,
    outdir: Path | None,
    tag: str,
    cmap: str,
    labels,
    add_colorbar: bool,
    nx: int,
    ny: int,
    slice: str,
):
    import matplotlib.pyplot as plt

    minval = np.min(img)
    maxval = np.max(img)
    fig, ax = plt.subplots(nx, ny, figsize=(20, 20))

    slice2axis = {"x": 0, "y": 1, "z": 2}
    axis = slice2axis[slice]

    for k, i in enumerate(np.linspace(0, img.shape[axis], 26, dtype=int)[:-1]):
        ax.flatten()[k].set_title(f"{i}")
        img_i = img.take(indices=i, axis=axis)
        im = ax.flatten()[k].imshow(img_i, cmap=cmap, vmin=minval, vmax=maxval)
    if add_colorbar:
        fig.subplots_adjust(right=0.8)
        cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
        cax = fig.colorbar(im, cax=cbar_ax)
        if labels:
            cax.ax.set_yticks(np.array(list(labels.values())))
            cax.ax.set_yticklabels(np.array(list(labels.keys())))
    if outdir is not None:
        fig.savefig(outdir / f"{tag}x_slice.png")
    else:
        plt.show()
    plt.close("all")


def plot_slices(
    img: np.ndarray,
    outdir: Path | None = None,
    tag="",
    cmap="gray",
    labels=None,
    add_colorbar=False,
    nx: int = 5,
    ny: int = 5,
    slice: typing.Literal["x", "y", "z", "all"] = "all",
):
    if slice == "all":
        plot_slice(img, outdir, tag, cmap, labels, add_colorbar, nx, ny, "x")
        plot_slice(img, outdir, tag, cmap, labels, add_colorbar, nx, ny, "y")
        plot_slice(img, outdir, tag, cmap, labels, add_colorbar, nx, ny, "z")
    else:
        plot_slice(img, outdir, tag, cmap, labels, add_colorbar, nx, ny, slice)



from pathlib import Path
import logging

logger = logging.getLogger(__name__)


def add_arguments(parser):
    parser.add_argument(
        "-o",
        "--outdir",
        type=Path,
        required=True,
        help="Output directory",
    )
    parser.add_argument(
        "--r",
        type=float,
        default=0.1,
        help="Radius of the brain",
    )
    parser.add_argument(
        "--parenchyma-factor",
        type=float,
        default=0.6,
        help="Parenchyma factor",
    )
    parser.add_argument(
        "--lv-factor",
        type=float,
        default=0.2,
        help="Left ventricle factor",
    )
    parser.add_argument(
        "--v34-center-factor",
        type=float,
        default=0.4,
        help="Ventricle 3/4 center factor",
    )
    parser.add_argument(
        "--v34-height-factor",
        type=float,
        default=0.42,
        help="Ventricle 3/4 height factor",
    )
    parser.add_argument(
        "--v34-radius-factor",
        type=float,
        default=0.06,
        help="Ventricle 3/4 radius factor",
    )
    parser.add_argument(
        "--skull-x0",
        type=float,
        default=0.0103891,
        help="Skull x0",
    )
    parser.add_argument(
        "--skull-x1",
        type=float,
        default=0.155952,
        help="Skull x1",
    )
    parser.add_argument(
        "--skull-y0",
        type=float,
        default=0.0114499,
        help="Skull y0",
    )
    parser.add_argument(
        "--skull-y1",
        type=float,
        default=0.173221,
        help="Skull y1",
    )
    parser.add_argument(
        "--skull-z0",
        type=float,
        default=0.001,
        help="Skull z0",
    )
    parser.add_argument(
        "--skull-z1",
        type=float,
        default=0.154949,
        help="Skull z1",
    )
    parser.add_argument(
        "--epsilon",
        type=float,
        default=0.0009,
        help="Epsilon",
    )
    parser.add_argument(
        "--edge-length-r",
        type=float,
        default=0.015,
        help="Edge length",
    )
    parser.add_argument(
        "--skip-simplify",
        action="store_true",
        help="Skip simplification",
    )
    parser.add_argument(
        "--coarsen",
        action="store_true",
        help="Coarsen",
    )
    parser.add_argument(
        "--stop-quality",
        type=int,
        default=8,
        help="Stop quality",
    )
    parser.add_argument(
        "--max-its",
        type=int,
        default=30,
        help="Max iterations",
    )
    parser.add_argument(
        "--loglevel",
        type=int,
        default=10,
        help="Log level",
    )
    parser.add_argument(
        "--disable-filtering",
        action="store_true",
        help="Disable filtering",
    )
    parser.add_argument(
        "--use-floodfill",
        action="store_true",
        help="Use floodfill",
    )
    parser.add_argument(
        "--smooth-open-boundary",
        action="store_true",
        help="Smooth open boundary",
    )
    parser.add_argument(
        "--manifold-surface",
        action="store_true",
        help="Manifold surface",
    )


def main(
    outdir: Path,
    r: float = 0.1,
    parenchyma_factor: float = 0.6,
    lv_factor: float = 0.2,
    v34_center_factor: float = 0.4,
    v34_height_factor: float = 0.42,
    v34_radius_factor: float = 0.06,
    skull_x0: float = 0.0103891,
    skull_x1: float = 0.155952,
    skull_y0: float = 0.0114499,
    skull_y1: float = 0.173221,
    skull_z0: float = 0.001,
    skull_z1: float = 0.154949,
    epsilon: float = 0.0009,
    edge_length_r: float = 0.015,
    skip_simplify: bool = False,
    coarsen: bool = True,
    stop_quality: int = 8,
    max_its: int = 30,
    loglevel: int = 10,
    disable_filtering: bool = False,
    use_floodfill: bool = False,
    smooth_open_boundary: bool = False,
    manifold_surface: bool = False,
) -> None:
    logger.info("Generating idealized brain surface")
    from ..surface.idealized_brain import main as main_surface

    main_surface(
        outdir,
        r=r,
        parenchyma_factor=parenchyma_factor,
        lv_factor=lv_factor,
        v34_center_factor=v34_center_factor,
        v34_height_factor=v34_height_factor,
        v34_radius_factor=v34_radius_factor,
        skull_x0=skull_x0,
        skull_x1=skull_x1,
        skull_y0=skull_y0,
        skull_y1=skull_y1,
        skull_z0=skull_z0,
        skull_z1=skull_z1,
    )
    from .basic import create_mesh, CSGTree, convert_mesh_dolfinx

    csg_tree: CSGTree = {
        "operation": "union",
        "right": {
            "operation": "union",
            "left": f"{outdir}/LV.ply",
            "right": f"{outdir}/V34.ply",
        },
        "left": {
            "operation": "union",
            "left": f"{outdir}/skull.ply",
            "right": f"{outdir}/parenchyma_incl_ventr.ply",
        },
    }

    create_mesh(
        outdir,
        csg_tree=csg_tree,
        epsilon=epsilon,
        edge_length_r=edge_length_r,
        skip_simplify=skip_simplify,
        coarsen=coarsen,
        stop_quality=stop_quality,
        max_its=max_its,
        loglevel=loglevel,
        disable_filtering=disable_filtering,
        use_floodfill=use_floodfill,
        smooth_open_boundary=smooth_open_boundary,
        manifold_surface=manifold_surface,
    )
    convert_mesh_dolfinx(mesh_dir=outdir)



import typing
import argparse
import logging

from . import parenchyma, idealized_brain

logger = logging.getLogger(__name__)


def add_surface_parser(parser: argparse.ArgumentParser) -> None:
    subparsers = parser.add_subparsers(
        dest="surface-command",
    )

    parenchyma_parser = subparsers.add_parser("parenchyma", help="Generate parenchyma surface")
    parenchyma.add_arguments(parenchyma_parser)

    idealized_parser = subparsers.add_parser(
        "idealized",
        help="Generate idealized surface",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    idealized_brain.add_arguments(idealized_parser)


def dispatch(command, args: dict[str, typing.Any]) -> int:
    if command == "parenchyma":
        parenchyma.main(**args)

    elif command == "idealized":
        idealized_brain.main(**args)

    else:
        raise ValueError(f"Unknown command {command}")

    return 0



import logging
import numpy as np
import numpy.typing as npt
import skimage.morphology as skim
import skimage


logger = logging.getLogger(__name__)


def get_closest_point(a: npt.NDArray, b: np.ndarray) -> tuple[np.int64, ...]:
    """Get the index of the closest point on `b` to `a`

    Parameters
    ----------
    a : npt.NDArray
        The object to find the closest point to
    b : np.ndarray
        The object to find the closest point on

    Returns
    -------
    tuple[int, ...]
        The index of the closest point

    Example
    -------

    >>> a = np.zeros((10, 10, 10), dtype=bool)
    >>> b = np.zeros((10, 10, 10), dtype=bool)
    >>> a[:2, :2, :2] = True
    >>> b[5:7, 5:7, 5:7] = True
    >>> a_b_index = get_closest_point(a, b)
    >>> assert np.allclose(a_b_index, (5, 5, 5))

    """
    import scipy.ndimage as ndi

    dist = ndi.distance_transform_edt(np.logical_not(a))
    assert isinstance(dist, np.ndarray)
    dist[np.logical_not(b)] = np.inf
    minidx = np.unravel_index(np.argmin(dist), a.shape)
    return minidx


def connect_by_line(
    m1: npt.NDArray, m2: npt.NDArray, footprint: npt.NDArray = skim.ball(1)
) -> npt.NDArray:
    """Connect two objects by a line

    Parameters
    ----------
    m1 : npt.NDArray
        The first object
    m2 : npt.NDArray
        The second object
    footprint : npt.NDArray, optional
        The footprint to use for dilation, by default skim.ball(1)

    Returns
    -------
    npt.NDArray
        The connected objects

    """
    # compute connection between V3 and V4:
    pointa = get_closest_point(m1, m2)
    pointb = get_closest_point(m2, m1)

    # add a line between the shortest points to connect V3 and V4
    line = np.array(skimage.draw.line_nd(pointa, pointb, endpoint=True))
    conn = np.zeros(m1.shape, dtype=np.uint8)
    i, j, k = line
    conn[i, j, k] = 1
    return skim.binary_dilation(conn, footprint=footprint)


def binary_smoothing(img: npt.NDArray, footprint=skim.ball(1)) -> npt.NDArray:
    openend = skim.binary_opening(img, footprint=footprint)
    return skim.binary_closing(openend, footprint=footprint)


def seperate_labels(
    img: npt.NDArray, l1: list[int], l2: list[int], dist: float, newlabel: int
) -> npt.NDArray:
    logger.debug(f"Seperating labels {l1} and {l2} with distance {dist}, new label {newlabel}")
    new_img = img.copy()
    m1 = skim.binary_dilation(img == l1, skim.ball(dist))
    m2 = skim.binary_dilation(img == l2, skim.ball(dist))
    new_img[np.logical_and(m1, m2)] = newlabel
    return new_img



import argparse
from pathlib import Path
import pyvista as pv


def add_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-i",
        "--input",
        type=Path,
        required=True,
        help="Path to the file to analyze",
    )


def volume_clip(mesh: pv.ImageData) -> None:
    p = pv.Plotter()
    p.add_mesh_clip_plane(mesh)
    p.show()


def main(input: Path) -> int:
    reader = pv.get_reader(input)
    mesh = reader.read()
    volume_clip(mesh)

    return 0



from __future__ import annotations
import argparse
import typing
from pathlib import Path

import numpy as np
import skimage.morphology as skim
import pyvista as pv

from ..constants import HOLE_THRESHOLD
from .utils import extract_surface
from ..viz import plot_slices
from ..constants import MM2M
from ..reader import read


def generate_mask(
    img, labels: dict[str, list[int]], hole_threshold: int = HOLE_THRESHOLD
) -> np.ndarray:
    assert "FLUID" in labels, "Need to provide FLUID label"
    par_mask = np.logical_not(np.isin(img, labels["FLUID"]))
    par_mask = skim.remove_small_objects(par_mask, hole_threshold)
    par_mask = skim.remove_small_holes(par_mask, hole_threshold)
    return par_mask


def generate_surface(
    outdir,
    img=None,
    resolution=(1, 1, 1),
    origin=(0, 0, 0),
    par_mask=None,
    labels: dict[str, list[int]] | None = None,
    plot_mask: bool = False,
):
    if par_mask is None:
        assert img is not None, "Need to provide image if par_mask is not provided"
        assert labels is not None, "Need to provide labels if par_mask is not provided"
        par_mask = generate_mask(img, labels)
    if plot_mask:
        plot_slices(par_mask, outdir, tag="parenchyma_mask_")
    par_surf = extract_surface(par_mask, resolution=resolution, origin=origin)
    par_surf = par_surf.smooth_taubin(n_iter=10, pass_band=0.1)
    # par_surf.points = nibabel.affines.apply_affine(seg.affine, par_surf.points)
    par_surf = par_surf.clip_closed_surface(normal=(0, 0, 1), origin=(0, 0, 1))
    par_surf.compute_normals(inplace=True, flip_normals=False)
    pv.save_meshio(f"{outdir}/parenchyma.ply", par_surf.scale(MM2M))


def add_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-i",
        "--input",
        type=Path,
        required=True,
        help="Path to the file to analyze",
    )
    parser.add_argument(
        "--input-robust",
        type=Path,
        help="Path to the robust segmentation to use as mask",
    )
    parser.add_argument(
        "-o",
        "--outdir",
        type=Path,
        default="results",
        help="Path to the output directory",
    )
    parser.add_argument(
        "--padding",
        type=int,
        default=5,
        help="Padding to add to the image",
    )
    parser.add_argument(
        "--labels",
        "-l",
        type=str,
        choices=["synthseg", "neurquant"],
        default="synthseg",
        help="Labels to use for the segmentation, default is synthseg",
    )
    parser.add_argument(
        "--plot-mask",
        action="store_true",
        help="Plot slices of the mask",
    )
    parser.add_argument(
        "--hemisphere-min-distance",
        type=float,
        default=0.0,
        help="Minimum distance between hemispheres",
    )


def main(
    input: Path,
    outdir: Path,
    padding: int,
    labels: typing.Literal["synthseg", "neuroquant"] = "synthseg",
    input_robust: Path | None = None,
    plot_mask: bool = False,
    hemisphere_min_distance: float = 0.0,
) -> int:
    volume = read(input, input_robust, label_name=labels, padding=padding)
    volume.hemisphere_min_distance = hemisphere_min_distance

    outdir.mkdir(parents=True, exist_ok=True)
    generate_surface(
        outdir,
        img=volume.img,
        resolution=volume.resolution,
        origin=volume.origin,
        labels=volume.labels,
        plot_mask=plot_mask,
    )
    return 0



import numpy as np
import skimage.morphology as skim

from mri2mesh import morphology as morph


def test_closest_point():
    a = np.zeros((10, 10, 10), dtype=bool)
    b = np.zeros((10, 10, 10), dtype=bool)

    a[:2, :2, :2] = True
    b[5:7, 5:7, 5:7] = True

    a_b_index = morph.get_closest_point(a, b)
    assert np.allclose(a_b_index, (5, 5, 5))

    b_a_index = morph.get_closest_point(b, a)
    assert np.allclose(b_a_index, (1, 1, 1))


def test_connect_by_line():
    m1 = np.zeros((3, 3, 3), dtype=bool)
    m2 = np.zeros((3, 3, 3), dtype=bool)

    m1[0, 0, 0] = True
    m2[2, 0, 0] = True

    conn = morph.connect_by_line(m1, m2, footprint=skim.ball(1))
    expected = np.zeros((3, 3, 3), dtype=np.bool)
    expected[:, 0, 0] = True
    expected[:, 1, 0] = True
    expected[:, 0, 1] = True

    assert np.allclose(conn, expected)



from importlib.metadata import metadata

from . import viz
from . import morphology
from . import cli
from . import vtk_utils
from . import surface
from . import mesh
from . import reader
from .reader import read

__all__ = ["viz", "morphology", "cli", "vtk_utils", "surface", "reader", "read", "mesh"]

meta = metadata("mri2mesh")
__version__ = meta["Version"]
__author__ = meta["Author-email"]
__license__ = meta["License"]
__email__ = meta["Author-email"]
__program_name__ = meta["Name"]



import argparse
from pathlib import Path
import pyvista as pv

index2title = {0: "x", 1: "y", 2: "z"}


def add_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-i",
        "--input",
        type=Path,
        required=True,
        help="Path to the file to analyze",
    )
    parser.add_argument(
        "-a",
        "--axis",
        type=int,
        default=0,
        help="Axis to slice along, 0=x, 1=y, 2=z 3=all (no slider)",
    )
    parser.add_argument(
        "--slider",
        action="store_true",
        help="Add a slider to the plot",
    )


def volume_slice_all(mesh: pv.ImageData) -> None:
    p = pv.Plotter(shape=(1, 3), border=False)
    p.add_mesh(mesh.outline())

    for axis in range(3):
        p.subplot(0, axis)
        p.add_title(f"{index2title[axis]}-coordinate")
        p.add_mesh(mesh.outline())
        normal = [0, 0, 0]
        normal[axis] = 1
        origin = list(mesh.origin)
        origin[axis] = mesh.dimensions[axis] // 2
        p.add_mesh_slice(
            mesh,
            name="slice",
            normal=normal,
            origin=origin,
        )

    p.show()


def volume_slice(mesh: pv.ImageData, axis=0, slider: bool = False) -> None:
    assert axis in [0, 1, 2], f"Invalid axis {axis}"

    plotter = pv.Plotter()
    plotter.add_mesh(mesh.outline())
    plotter.add_title(f"{index2title[axis]}-coordinate")
    origin = list(mesh.origin)
    normal = [0, 0, 0]
    normal[axis] = 1
    origin[axis] = mesh.dimensions[axis] // 2

    if slider:

        def update(value):
            origin[axis] = value
            plotter.add_mesh_slice(mesh, name="slice", normal=normal, origin=origin)

        plotter.add_slider_widget(
            update, [0, mesh.dimensions[axis]], title=f"{index2title[axis]}-coordinate"
        )

    else:
        plotter.add_mesh_slice(
            mesh,
            name="slice",
            normal=normal,
            origin=origin,
        )

    plotter.show()


def main(input: Path, axis: int = 3, slider: bool = False) -> int:
    reader = pv.get_reader(input)
    mesh = reader.read()
    if axis == 3:
        assert not slider, "Slider not supported for all axes"
        volume_slice_all(mesh)
    else:
        volume_slice(mesh, axis=axis, slider=slider)

    return 0



import logging

logger = logging.getLogger(__name__)


def dispatch(name: str) -> None:
    logger.info(f"Listing segmentation labels for {name}")
    if name == "synthseg":
        for label, values in SYNTHSEG_LABELS.items():
            print(f"{label}: {values}")
    elif name == "neuroquant":
        for label, values in NEUROQUANT_LABELS.items():
            print(f"{label}: {values}")
    else:
        raise ValueError(f"Unknown segmentation labels {name}")


SYNTHSEG_LABELS = {
    "BACKGROUND": [0],
    "LEFT_CEREBRAL_WHITE_MATTER": [2],
    "LEFT_CEREBRAL_CORTEX": [3],
    "LEFT_LATERAL_VENTRICLE": [4],
    "LEFT_INFERIOR_LATERAL_VENTRICLE": [5],
    "LEFT_CEREBELLUM_WHITE_MATTER": [7],
    "LEFT_CEREBELLUM_CORTEX": [8],
    "LEFT_THALAMUS": [10],
    "LEFT_CAUDATE": [11],
    "LEFT_PUTAMEN": [12],
    "LEFT_PALLIDUM": [13],
    "THIRD_VENTRICLE": [14],
    "FOURTH_VENTRICLE": [15],
    "BRAIN_STEM": [16],
    "LEFT_HIPPOCAMPUS": [17],
    "LEFT_AMYGDALA": [18],
    "LEFT_ACCUMBENS_AREA": [26],
    "CSF": [24],
    "LEFT_VENTRAL_DC": [28],
    "RIGHT_CEREBRAL_WHITE_MATTER": [41],
    "RIGHT_CEREBRAL_CORTEX": [42],
    "RIGHT_LATERAL_VENTRICLE": [43],
    "RIGHT_INFERIOR_LATERAL_VENTRICLE": [44],
    "RIGHT_CEREBELLUM_WHITE_MATTER": [46],
    "RIGHT_CEREBELLUM_CORTEX": [47],
    "RIGHT_THALAMUS": [49],
    "RIGHT_CAUDATE": [50],
    "RIGHT_PUTAMEN": [51],
    "RIGHT_PALLIDUM": [52],
    "RIGHT_HIPPOCAMPUS": [53],
    "RIGHT_AMYGDALA": [54],
    "RIGHT_ACCUMBENS_AREA": [58],
    "RIGHT_VENTRAL_DC": [60],
    "CEREBRUM": [2, 41, 3, 42],
    "FLUID": [24, 14, 15, 4, 43, 5, 44, 0],
    "VENTRICLE": [14, 15, 4, 43, 5, 44],
    "CEREBRAL_WHITE_MATTER": [2, 41],
    "CEREBRAL_CORTEX": [3, 42],
    "CEREBELLUM_WHITE_MATTER": [7, 46],
    "CEREBELLUM_CORTEX": [8, 47],
    "CEREBELLUM_GRAY_MATTER": [8, 47],
}


NEUROQUANT_LABELS = {
    "BACKGROUND": [0],
    "CENTRAL_WHITE_MATTER": [1],
    "CORTICAL_GRAY_MATTER": [2],
    "THIRD_VENTRICLE": [3],
    "FOURTH_VENTRICLE": [4],
    "FIFTH_VENTRICLE": [5],
    "LATERAL_VENTRICLE": [6],
    "INF_LAT_VENTRICLE": [7],
    "LAT_VENTRICLE_CHOROID_PLEXUS": [8],
    "CEREBELLUM_WHITE_MATTER": [9],
    "CEREBELLUM_GRAY_MATTER": [10],
    "HIPPOCAMPUS": [11],
    "AMYGDALA": [12],
    "THALAMUS": [13],
    "CAUDATE": [14],
    "PUTAMEN": [15],
    "PALIDUM": [16],
    "VENTRAL_DC": [17],
    "NUCLEUS_ACCUMBENS": [18],
    "BRAIN_STEM": [19],
    "INF_LAT_VENTRICLE_CHOROID_PLEXUS": [20],
    "CEREBRAL_WHITE_MATTER_HYPOINTENSITY": [21],
    "EXTERIOR": [22],
    "UNKNOWN": [23],
    "POST_SUP_TEMPORAL_SULCUS": [24],
    "CAUD_ANT_CINGULATE": [25],
    "PREMOTOR": [26],
    "CORPUS_CALLOSUM": [27],
    "CUNEUS_CORTEX": [28],
    "ENTORHINAL_CORTEX": [29],
    "FUSIFORM_CORTEX": [30],
    "INF_PARIETAL_CORTEX": [31],
    "INF_TEMPORAL_CORTEX": [32],
    "ISTHMUS_CINGULATE": [33],
    "LAT_OCCIPITAL": [34],
    "LAT_ORBITOFRONTAL_GYRUS": [35],
    "LINGUAL_GYRUS": [36],
    "MED_ORBITOFRONTAL_GYRUS": [37],
    "MID_TEMPORAL_GYRUS": [38],
    "PARAHIPPOCAMPAL_GYRUS": [39],
    "PARACENTRAL": [40],
    "PARS_OPERCULARIS": [41],
    "PARS_ORBITALIS": [42],
    "PARS_TRIANGULARIS": [43],
    "PERICALCARINE": [44],
    "PRI_SENSORY_CORTEX": [45],
    "POST_CINGULATE_GYRUS": [46],
    "PRI_MOTOR_CORTEX": [47],
    "MED_PARIETAL_CORTEX": [48],
    "ROST_ANT_CINGULATE": [49],
    "ANT_MID_FRONTAL": [50],
    "SUP_FRONTAL_GYRUS": [51],
    "SUP_PARIETAL_LOBULE": [52],
    "SUP_TEMPORAL": [53],
    "SUPRAMARGINAL_GYRUS": [54],
    "FRONTAL_POLE": [55],
    "TEMPORAL_POLE": [56],
    "TRANSVERSE_TEMPORAL": [57],
    "INSULA": [58],
    "FLUID": [3, 4, 5, 6, 7, 8, 20],
    "VENTRICLE": [3, 4, 5, 6, 7, 8],
}



import typing
import json
import logging
import pprint
from pathlib import Path
import numpy as np

logger = logging.getLogger(__name__)


def sample_config():
    tree = {
        "operation": "union",
        "right": {
            "operation": "union",
            "left": "surfaces/LV.ply",
            "right": "surfaces/V34.ply",
        },
        "left": {
            "operation": "union",
            "left": "surfaces/skull.ply",
            "right": "surfaces/parenchyma_incl_ventr.ply",
        },
    }

    config = {
        "outdir": "mesh",
        "epsilon": 0.0009,
        "edge_length_r": 0.015,
        "coarsen": True,
        "stop_quality": 8,
        "max_its": 30,
        "loglevel": 10,
        "disable_filtering": False,
        "use_floodfill": False,
        "smooth_open_boundary": False,
        "manifold_surface": False,
        "csg_tree": tree,
    }
    return config


def default_config() -> dict[str, typing.Any]:
    config = sample_config()
    config["csg_tree"] = {}
    return config


def generate_sameple_config(outdir: Path, name: str = "config.json"):
    logger.info("Generating sample configuration file")
    config = sample_config()
    path = outdir / name
    if path.suffix == ".json":
        path.write_text(json.dumps(config, indent=2))
    elif path.suffix == ".toml":
        import toml

        path.write_text(toml.dumps(config))
    elif path.suffix == ".yaml":
        import yaml

        path.write_text(yaml.dump(config))
    else:
        raise ValueError(f"Unknown file extension {path.suffix}")
    logger.info("Configuration file written to %s", path)


def read_config(path: Path) -> dict[str, typing.Any]:
    logger.info("Reading configuration file %s", path)
    if path.suffix == ".json":
        return json.loads(path.read_text())
    elif path.suffix == ".toml":
        import toml

        return toml.loads(path.read_text())
    elif path.suffix == ".yaml":
        import yaml

        return yaml.load(path.read_text(), Loader=yaml.FullLoader)
    else:
        raise ValueError(f"Unknown file extension {path.suffix}")


class CSGTree(typing.TypedDict):
    operation: str
    left: typing.Union[str, "CSGTree"]
    right: typing.Union[str, "CSGTree"]


def create_mesh_from_config(filename: Path) -> Path:
    if not filename.exists():
        raise FileNotFoundError(f"File {filename} not found")
    config = default_config()
    config.update(read_config(filename))
    create_mesh(**config)
    return Path(config["outdir"])


def create_mesh(
    outdir: typing.Union[Path, str],
    csg_tree: CSGTree,
    epsilon: float = 0.0009,
    edge_length_r: float = 0.015,
    skip_simplify: bool = False,
    coarsen: bool = True,
    stop_quality: int = 8,
    max_its: int = 30,
    loglevel: int = 10,
    disable_filtering: bool = False,
    use_floodfill: bool = False,
    smooth_open_boundary: bool = False,
    manifold_surface: bool = False,
):
    logger.info("Creating mesh, with the following configuration")
    params = {
        "outdir": str(outdir),
        "csg_tree": csg_tree,
        "epsilon": epsilon,
        "edge_length_r": edge_length_r,
        "skip_simplify": skip_simplify,
        "coarsen": coarsen,
        "stop_quality": stop_quality,
        "max_its": max_its,
        "loglevel": loglevel,
        "disable_filtering": disable_filtering,
        "use_floodfill": use_floodfill,
        "smooth_open_boundary": smooth_open_boundary,
        "manifold_surface": manifold_surface,
    }
    logger.info(pprint.pformat(params))

    import wildmeshing as wm
    import meshio
    import pyvista as pv

    outdir = Path(outdir)
    outdir.mkdir(parents=True, exist_ok=True)
    (outdir / "mesh_params.json").write_text(json.dumps(params, indent=2))

    tetra = wm.Tetrahedralizer(
        epsilon=epsilon,
        edge_length_r=edge_length_r,
        coarsen=coarsen,
        stop_quality=stop_quality,
        max_its=max_its,
        skip_simplify=skip_simplify,
    )
    tetra.set_log_level(loglevel)

    tetra.load_csg_tree(json.dumps(csg_tree))
    tetra.tetrahedralize()
    point_array, cell_array, marker = tetra.get_tet_mesh(
        all_mesh=disable_filtering,
        smooth_open_boundary=smooth_open_boundary,
        floodfill=use_floodfill,
        manifold_surface=manifold_surface,
        correct_surface_orientation=True,
    )
    coords, connections = tetra.get_tracked_surfaces()

    tetra_mesh = meshio.Mesh(
        point_array, [("tetra", cell_array)], cell_data={"cell_tags": [marker.ravel()]}
    )
    tetra_mesh_pv = pv.from_meshio(tetra_mesh).clean()
    pv.save_meshio(outdir / "tetra_mesh.xdmf", tetra_mesh_pv)

    for i, coord in enumerate(coords):
        np.save(outdir / f"coords_{i}.npy", coord)

    for i, conn in enumerate(connections):
        np.save(outdir / f"connections_{i}.npy", conn)


def convert_mesh_dolfinx(mesh_dir: Path):
    logger.info("Converting mesh to dolfinx in %s", mesh_dir)
    from mpi4py import MPI
    from scipy.spatial.distance import cdist
    import dolfinx

    threshold = 1.0
    fdim = 2

    coords = []
    for path in sorted(mesh_dir.glob("coords_*.npy"), key=lambda x: int(x.stem.split("_")[-1])):
        coords.append(np.load(path))
    logger.debug(f"Found {len(coords)} coordinates")

    connections = []
    for path in sorted(
        mesh_dir.glob("connections_*.npy"), key=lambda x: int(x.stem.split("_")[-1])
    ):
        connections.append(np.load(path))
    logger.debug(f"Found {len(connections)} connections")

    assert len(connections) == len(coords)

    logger.debug("Loading mesh")
    comm = MPI.COMM_WORLD
    with dolfinx.io.XDMFFile(comm, mesh_dir / "tetra_mesh.xdmf", "r") as xdmf:
        mesh = xdmf.read_mesh(name="Grid")
        cell_tags = xdmf.read_meshtags(mesh, name="Grid")

    logger.debug("Mesh loaded")

    facets = []
    values = []
    for i, coord in enumerate(coords, start=1):
        logger.debug(f"Processing coord {i}")

        def locator(x):
            # Find the distance to all coordinates
            distances = cdist(x.T, coord)
            # And return True is they are close
            return np.any(distances < threshold, axis=1)

        f = dolfinx.mesh.locate_entities_boundary(mesh, dim=fdim, marker=locator)
        v = np.full(f.shape[0], i, dtype=np.int32)
        facets.append(f)
        values.append(v)

    logger.debug("Create meshtags")
    facet_tags = dolfinx.mesh.meshtags(
        mesh,
        fdim,
        np.hstack(facets),
        np.hstack(values),
    )
    facet_tags.name = "facet_tags"
    cell_tags.name = "cell_tags"
    mesh.name = "mesh"

    logger.debug("Save files")
    with dolfinx.io.XDMFFile(comm, mesh_dir / "mesh.xdmf", "w") as xdmf:
        xdmf.write_mesh(mesh)
        xdmf.write_meshtags(facet_tags, mesh.geometry)
        xdmf.write_meshtags(cell_tags, mesh.geometry)

    logger.info("Mesh saved to %s", mesh_dir / "mesh.xdmf")



import logging
import argparse
from typing import Sequence, Optional

from . import viz, surface, mesh, segmentation_labels


def setup_parser():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    # Root parser
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Just print the command and do not run it",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Print more information",
    )
    subparsers = parser.add_subparsers(dest="command")

    # Visualization parser
    viz_parser = subparsers.add_parser("viz", help="Visualize data")
    viz.add_viz_parser(viz_parser)

    # Surface generation parser
    surface_parser = subparsers.add_parser("surface", help="Generate surfaces")
    surface.add_surface_parser(surface_parser)

    # Mesh generation parser
    mesh_parser = subparsers.add_parser("mesh", help="Generate meshes")
    mesh.add_mesh_parser(mesh_parser)

    label_parser = subparsers.add_parser("labels", help="List segmentation labels")
    label_parser.add_argument(
        "name", type=str, choices=["synthseg", "neuroquant"], help="Name of the labels to display"
    )

    return parser


def _disable_loggers():
    for libname in ["matplotlib"]:
        logging.getLogger(libname).setLevel(logging.WARNING)


def dispatch(parser: argparse.ArgumentParser, argv: Optional[Sequence[str]] = None) -> int:
    args = vars(parser.parse_args(argv))
    logging.basicConfig(level=logging.DEBUG if args.pop("verbose") else logging.INFO)
    _disable_loggers()

    logger = logging.getLogger(__name__)
    dry_run = args.pop("dry_run")
    command = args.pop("command")

    if dry_run:
        logger.info("Dry run: %s", command)
        logger.info("Arguments: %s", args)
        return 0

    try:
        if command == "viz":
            viz.dispatch(args.pop("viz-command"), args)
        elif command == "surface":
            surface.dispatch(args.pop("surface-command"), args)
        elif command == "mesh":
            mesh.dispatch(args.pop("mesh-command"), args)
        elif command == "labels":
            segmentation_labels.dispatch(args.pop("name"))
        else:
            logger.error(f"Unknown command {command}")
            parser.print_help()
    except ValueError as e:
        logger.error(e)
        parser.print_help()

    return 0


def main(argv: Optional[Sequence[str]] = None) -> int:
    parser = setup_parser()
    return dispatch(parser, argv)



import pytest

from mri2mesh import cli


@pytest.mark.parametrize("name", ["synthseg", "neuroquant"])
def test_labels_cli(name, capsys):
    cli.main(["labels", name])
    captured = capsys.readouterr()
    # Both labels have Background as the first label
    assert captured.out.startswith("BACKGROUND: [0]")


def test_mesh_idealized_cli(tmp_path):
    cli.main(["mesh", "idealized", "-o", str(tmp_path)])
    for name in [
        "connections_3.npy",
        "coords_0.npy",
        "coords_1.npy",
        "connections_2.npy",
        "connections_0.npy",
        "coords_3.npy",
        "coords_2.npy",
        "connections_1.npy",
        "mesh.h5",
        "ventricles.ply",
        "skull.ply",
        "tetra_mesh.xdmf",
        "mesh.xdmf",
        "LV.ply",
        "surface_parameters.json",
        "tetra_mesh.h5",
        "parenchyma_incl_ventr.ply",
        "V34.ply",
        "mesh_params.json",
    ]:
        assert (tmp_path / name).exists()



from pathlib import Path
import argparse
import logging
import json
import datetime
import numpy as np

logger = logging.getLogger(__name__)


def add_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-o",
        "--outdir",
        type=Path,
        required=True,
        help="Output directory",
    )
    parser.add_argument(
        "--r",
        type=float,
        default=0.1,
        help="Radius of the brain",
    )
    parser.add_argument(
        "--parenchyma-factor",
        type=float,
        default=0.6,
        help="Parenchyma factor",
    )
    parser.add_argument(
        "--lv-factor",
        type=float,
        default=0.2,
        help="Left ventricle factor",
    )
    parser.add_argument(
        "--v34-center-factor",
        type=float,
        default=0.4,
        help="Ventricle 3/4 center factor",
    )
    parser.add_argument(
        "--v34-height-factor",
        type=float,
        default=0.42,
        help="Ventricle 3/4 height factor",
    )
    parser.add_argument(
        "--v34-radius-factor",
        type=float,
        default=0.06,
        help="Ventricle 3/4 radius factor",
    )
    parser.add_argument(
        "--skull-x0",
        type=float,
        default=0.0103891,
        help="Skull x0",
    )
    parser.add_argument(
        "--skull-x1",
        type=float,
        default=0.155952,
        help="Skull x1",
    )
    parser.add_argument(
        "--skull-y0",
        type=float,
        default=0.0114499,
        help="Skull y0",
    )
    parser.add_argument(
        "--skull-y1",
        type=float,
        default=0.173221,
        help="Skull y1",
    )
    parser.add_argument(
        "--skull-z0",
        type=float,
        default=0.001,
        help="Skull z0",
    )
    parser.add_argument(
        "--skull-z1",
        type=float,
        default=0.154949,
        help="Skull z1",
    )


def main(
    outdir: Path,
    r: float = 0.1,
    parenchyma_factor: float = 0.6,
    lv_factor: float = 0.2,
    v34_center_factor: float = 0.4,
    v34_height_factor: float = 0.42,
    v34_radius_factor: float = 0.06,
    skull_x0: float = 0.0103891,
    skull_x1: float = 0.155952,
    skull_y0: float = 0.0114499,
    skull_y1: float = 0.173221,
    skull_z0: float = 0.001,
    skull_z1: float = 0.154949,
) -> None:
    import pyvista as pv

    logger.info("Creating idealized brain surface in %s", outdir)
    outdir.mkdir(exist_ok=True, parents=True)

    z = np.array([0, 0, 1])
    skull = pv.Box((skull_x0, skull_x1, skull_y0, skull_y1, skull_z0, skull_z1))
    c = skull.center_of_mass()
    par = pv.Sphere(r * parenchyma_factor, center=c)
    LV = pv.Sphere(r * lv_factor, center=c)
    V34 = pv.Cylinder(
        c - v34_center_factor * r * z,
        direction=z,
        height=r * v34_height_factor,
        radius=v34_radius_factor * r,
    )
    ventricles = pv.merge([LV, V34])

    for s, n in zip(
        [V34, LV, par, skull, ventricles],
        ["V34", "LV", "parenchyma_incl_ventr", "skull", "ventricles"],
    ):
        pv.save_meshio(outdir / f"{n}.ply", s)

    from .. import __version__

    (outdir / "surface_parameters.json").write_text(
        json.dumps(
            {
                "r": r,
                "parenchyma_factor": parenchyma_factor,
                "lv_factor": lv_factor,
                "v34_center_factor": v34_center_factor,
                "v34_height_factor": v34_height_factor,
                "v34_radius_factor": v34_radius_factor,
                "skull_x0": skull_x0,
                "skull_x1": skull_x1,
                "skull_y0": skull_y0,
                "skull_y1": skull_y1,
                "skull_z0": skull_z0,
                "skull_z1": skull_z1,
                "version": __version__,
                "timestamp": datetime.datetime.now().isoformat(),
            },
            indent=2,
        )
    )



import numpy as np
import pyvista as pv


def numpy_to_vtkImageData(img: np.ndarray) -> pv.ImageData:
    import vtk
    import vtk.util.numpy_support as numpy_support

    if img.dtype in (np.float64, np.float32):
        data_type = vtk.VTK_FLOAT
    elif img.dtype in (np.int64, np.int32):
        data_type = vtk.VTK_INT
    else:
        raise ValueError(f"Unsupported data type: {img.dtype}")

    shape = img.shape[::-1]

    flat_img_array = img.ravel()
    vtk_data = numpy_support.numpy_to_vtk(num_array=flat_img_array, deep=True, array_type=data_type)

    vtk_img = vtk.vtkImageData()
    vtk_img.GetPointData().SetScalars(vtk_data)
    vtk_img.SetDimensions(*shape)
    return pv.ImageData(vtk_img)



from __future__ import annotations
from pathlib import Path
import logging
import typing
import numpy as np
import numpy.typing as npt

from dataclasses import dataclass, field

from .segmentation_labels import NEUROQUANT_LABELS, SYNTHSEG_LABELS

logger = logging.getLogger(__name__)


@dataclass  # (slots=True) # - will be added when we drop python3.9
class Segmentation:
    img: np.ndarray
    labels: dict[str, list[int]] = field(default_factory=lambda: SYNTHSEG_LABELS)
    resolution: npt.NDArray[np.float64] = field(default_factory=lambda: np.array((1, 1, 1)))
    origin: npt.NDArray[np.float64] = field(default_factory=lambda: np.array((0, 0, 0)))
    padding: int = 0
    _hemisphere_min_distance: float = 0.0

    def add_padding(self, padding: int) -> None:
        self.img = np.pad(self.img, padding)
        self.origin = -np.array(self.resolution) * padding

    @property
    def hemisphere_min_distance(self) -> float:
        return self._hemisphere_min_distance

    @hemisphere_min_distance.setter
    def hemisphere_min_distance(self, min_distance: float) -> None:
        if (
            min_distance > 0
            and "LEFT_CEREBRAL_CORTEX" in self.labels
            and "RIGHT_CEREBRAL_CORTEX" in self.labels
            and "CSF" in self.labels
        ):
            from .morphology import seperate_labels

            l1 = self.labels["LEFT_CEREBRAL_CORTEX"]
            l2 = self.labels["RIGHT_CEREBRAL_CORTEX"]
            self.img = seperate_labels(self.img, l1, l2, min_distance, self.labels["CSF"][0])

        self._hemisphere_min_distance = min_distance


def read(
    input: Path,
    input_robust: Path | None = None,
    label_name: typing.Literal["synthseg", "neuroquant"] = "synthseg",
    padding: int = 5,
) -> Segmentation:
    import nibabel as nib

    logger.info(f"Loading segmentation from {input}")
    seg = nib.load(input)
    img = np.pad(seg.get_fdata(), padding)
    logger.debug(f"Loaded segmentation with shape {img.shape} and padding {padding}")

    if label_name == "synthseg":
        labels = SYNTHSEG_LABELS
    elif label_name == "neuroquant":
        labels = NEUROQUANT_LABELS
    else:
        raise ValueError(f"Invalid labels {label_name}")

    if input_robust:
        img_rob = np.pad(nib.load(input_robust).get_fdata(), padding)

        img[img_rob == 0] = 0
        img[img_rob == labels["BRAIN_STEM"]] = labels["BRAIN_STEM"]
    resolution: npt.NDArray[np.float64] = np.array(seg.header["pixdim"][1:4])
    origin: npt.NDArray[np.float64] = -np.array(resolution) * padding
    return Segmentation(
        img=img,
        labels=labels,
        resolution=resolution,
        origin=origin,
        padding=padding,
    )
